#ifndef _AGGRESSIVEDISCARDHEAP_H_
#define _AGGRESSIVEDISCARDHEAP_H_

#include <sys/mman.h>

#include "pageinfo.h"
#include "util.h"

namespace HL {

//class AggressiveDiscardHeap: aggressively discard free page(s) in the object being freed
//note it is OK to discard the same page multiple times
template<class SuperHeap>
class AggressiveDiscardHeap : public SuperHeap {
  public:

	inline void free(void * ptr) {
		size_t size = getSize(ptr);

		size_t discard_start_page = (reinterpret_cast<size_t>(ptr) + 8 + PAGE_SIZE - 1) & PAGE_MASK;	//we know that there's 8 bytes used for free list
		size_t discard_end_page = (reinterpret_cast<size_t>(ptr) + size) & PAGE_MASK;

		if (discard_end_page > discard_start_page) {
			dbprintf("discarding pages from %p to %p\n", discard_start_page, discard_end_page);

			int rc = madvise(reinterpret_cast<void *>(discard_start_page), discard_end_page - discard_start_page, MADV_DONTNEED);
			assert(rc == 0);
		}

		SuperHeap::free(ptr);
	}

};	//end of class AggressiveDiscardHeap

};	//end of namespace HL

#endif
