#ifndef _BFHEAP_H_
#define _BFHEAP_H_

#include "dllisttype.h"

namespace HL {

//class FFHeap: a heap using the best-fit policy
template<int Type, class SuperHeap>
class BFHeap : public SuperHeap {
  public:
	inline void * malloc(size_t size) {
		void * ptr = free_list.getListNext(NULL);
		size_t best_size = 0xffffffff;
		void * best_fit = NULL;

		while (ptr != NULL) {
			if (getSize(ptr) >= size && getSize(ptr) < best_size) {
				best_size = getSize(ptr);
				best_fit = ptr;
			}

			ptr = free_list.getListNext(reinterpret_cast<DLListEntry *>(ptr));
		}

		if (best_fit == NULL) {
			ptr = SuperHeap::malloc(size);
			assert(ptr != NULL);
		}
		else {
			ptr = best_fit;
			assert(getSize(ptr) >= size);
			free_list.remove(reinterpret_cast<DLListEntry *>(ptr));
		}

		return ptr;
	}

	inline void free(void * ptr) {
		free_list.insert(reinterpret_cast<DLListEntry *>(ptr));
	}

	inline void remove(void * ptr) {
		free_list.remove(reinterpret_cast<DLListEntry *>(ptr));
	}

  private:
	DLListType<Type> free_list;

};	//end of class BFHeap
};	//end of namespace HL

#endif
