#ifndef _DLLISTTYPE_H_
#define _DLLISTTYPE_H_

#include "util.h"

#define FIFO	1
#define LIFO	2
#define AO		3

namespace HL {

struct DLListEntry {
	DLListEntry * prev;
	DLListEntry * next;

	inline void insert(DLListEntry * p, DLListEntry * n) {
		prev = p;
		next = n;
		p->next = this;
		n->prev = this;
	}
	inline void remove() {
		prev->next = next;
		next->prev = prev;
	}
};


//class DLListType: a doubly-linked list that works in FIFO, LIFO or AO order
template<int Type>
class DLListType {
  public:

	inline DLListType() {
		assert(Type == FIFO || Type == LIFO || Type == AO);
		clear();

		sanityCheck();
	}

	inline void clear() {
		head.prev = &head;
		head.next = &head;
		num_entries = 0;

		ao_last_insert = &head;
	}
/*
	inline DLListEntry * get() {
		const DLListEntry * e = head.next;
		if (e == &head) {
			return NULL;
		}
		head.next = e->next;
		head.next->prev = &head;
		return (DLListEntry *) e;
	}
*/

	inline DLListEntry * getListPrev(DLListEntry * current) {
		sanityCheck();

		assert(current != NULL);
		return current->prev;
	}

	//get the next entry from the list
	//but don't remove this entry
	inline DLListEntry * getListNext(DLListEntry * current) {
		sanityCheck();

		if (current == NULL) {
			current = &head;
		}
		else {
			checkEntry(current);
		}

		if (current->next != &head) {
			return current->next;
		}
		else {
			return NULL;
		}
	}

	inline void remove(DLListEntry * e) {
		sanityCheck();
		checkEntry(e);

		if (ao_last_insert == e) {
			ao_last_insert = e->prev;
		}

		e->remove();
		num_entries--;

		sanityCheck();
	}

	inline void insert(DLListEntry * e) {
		sanityCheck();

		switch (Type) {
		case FIFO:						//FIFO: insert into the end
			e->insert(head.prev, &head);
			break;
		case LIFO:						//LIFO: insert into the front
			e->insert(&head, head.next);
			break;
		case AO:						//AO: find the right place to insert
		{
#if 0
			DLListEntry * i = head.next;
			int n = 0;
			while (1) {		//search from the front towards the end (inefficient)
				assert(i != e);
				if (i > e || i == &head) {
					e->insert(i->prev, i);
					break;
				}
				i = i->next;
				n++;
				assert(n <= num_entries);
			}
#else
			assert(ao_last_insert != e);

			if (head.next == &head) {	//list is currently empty
				assert(num_entries == 0);
				assert(ao_last_insert == &head);
				e->insert(&head, head.next);
				ao_last_insert = e;
				break;
			}

			assert(num_entries > 0);

			if (ao_last_insert == &head) {
				ao_last_insert = head.next;
			}

			checkEntry(ao_last_insert);

			if (e < ao_last_insert) {	//we should go towards the front
				int n = 0;
				while (1) {
					assert(ao_last_insert != e);
					if (ao_last_insert < e || ao_last_insert == &head) {
						e->insert(ao_last_insert, ao_last_insert->next);
						ao_last_insert = e;
						break;
					}
					ao_last_insert = ao_last_insert->prev;
					assert(n <= num_entries);
				}
			}
			else {						//we should go towards the end
				int n = 0;
				while (1) {
					assert(ao_last_insert != e);
					if (ao_last_insert > e || ao_last_insert == &head) {
						e->insert(ao_last_insert->prev, ao_last_insert);
						ao_last_insert = e;
						break;
					}
					ao_last_insert = ao_last_insert->prev;
					assert(n <= num_entries);
				}
			}

			assert(ao_last_insert == e);
#endif
			break;
		}
		default:
			assert(false);
			break;
		}
		num_entries++;

		sanityCheck();
		checkEntry(e);
	}

  private:
	DLListEntry head;
	int num_entries;

	DLListEntry * ao_last_insert;	//for AO only, a cached position of the last insert

	void checkEntry(DLListEntry * e) {
#ifdef DEBUG
#ifdef SANITY_CHECK
		assert(e != NULL);
		assert(e != &head);

		DLListEntry * s = head.next;
		int n = num_entries;

		while (s != &head) {
			n--;
			assert(n >= 0);
			if (s == e) {
				break;
			}
			s = s->next;
		}

		assert(s == e);
#endif
#endif
	}

	void sanityCheck() {
#ifdef DEBUG
#ifdef SANITY_CHECK
		DLListEntry * e = head.next;
		int n = 0;

		while (e != &head) {
			assert(e != NULL);
			assert(e->prev->next == e);
			assert(e->next->prev == e);
			e = e->next;
			n++;
		}

		assert(n == num_entries);
#endif
#endif
	}

};	//end of class DLListType
};	//end of namespace HL

#endif
