#ifndef _ENHANCEDSBRKHEAP_H_
#define _ENHANCEDSBRKHEAP_H_

#include "util.h"

//class EnhancedSbrkHeap: differs from Sbrk in
//1: initial break is aligned
//2: "slop" is integrated into this heap
//3: 
template<int Slop>
class EnhancedSbrkHeap {
  public:
	inline EnhancedSbrkHeap() {
		heap_end = sbrk(0);

		//align initial break (heap end)
		if (reinterpret_cast<size_t>(heap_end) % sizeof(double) != 0) {
			heap_end = reinterpret_cast<void *>((reinterpret_cast<size_t>(heap_end) + sizeof(double) - 1) / sizeof(double) * sizeof(double));
			int rc = brk(heap_end);
			assert(rc == 0);
		}

		assert(reinterpret_cast<size_t>(sbrk(0)) % sizeof(double) == 0);

		assert(Slop >= 0);
		sbrk(Slop);

		assert(reinterpret_cast<size_t>(sbrk(0)) == reinterpret_cast<size_t>(heap_end) + Slop);
	}

	inline void * malloc(size_t size) {
		assert(size > 0 && size % sizeof(double) == 0);
		void * ptr = heap_end;

		heap_end = sbrk(size);
		dbprintf("heap_end=%p ptr=%p Slop=%d\n", heap_end, ptr, Slop);
		assert(reinterpret_cast<size_t>(heap_end) == reinterpret_cast<size_t>(ptr) + Slop);

		heap_end = reinterpret_cast<void *>(reinterpret_cast<size_t>(ptr) + size);

		assert(reinterpret_cast<size_t>(sbrk(0)) == reinterpret_cast<size_t>(heap_end) + Slop);
		return ptr;
	}

	inline int shrink_brk(void * ptr, size_t size) {
		assert(reinterpret_cast<size_t>(ptr) + size <= reinterpret_cast<size_t>(heap_end));

		if (reinterpret_cast<size_t>(ptr) + size == reinterpret_cast<size_t>(heap_end)) {
			heap_end = ptr;
			sbrk(-size);
			assert(reinterpret_cast<size_t>(sbrk(0)) == reinterpret_cast<size_t>(heap_end) + Slop);
			return 1;
		}
		else {
			return 0;
		}
	}

  private:

	void * heap_end;

};	//end of class EnhancedSbrkHeap

#endif
