#ifndef _LIB_H_
#define _LIB_H_

#include "heaplayers.h"
#include "slopheap.h"
#include "chunkheap.h"
#include "mallocheap.h"

#define DB_PRINT_TO_FILE	1

#ifndef USE_LOGHEAP
#define USE_LOGHEAP			1
#endif

#ifndef DISCARD_FREE_PAGES
#define DISCARD_FREE_PAGES	1
#endif

#include "timeintervalheap.h"

#include "dllisttype.h"
#include "ffheap.h"
#include "nfheap.h"
#include "bfheap.h"
#include "enhancedsbrkheap.h"
#include "shrinkbrkheap.h"
#include "logheap.h"


#if DISCARD_FREE_PAGES
#include "aggressivediscardheap.h"
#else
template<class SuperHeap>
class AggressiveDiscardHeap : public SuperHeap {};
#endif

template<class SuperHeap>
class CommonWrapper :
public ANSIWrapper<LogHeap<SuperHeap> > {};
//public LogHeap<ANSIWrapper<SuperHeap> > {};

class SeqFitCommonBaseHeap :
public CoalesceableHeap<EnhancedSbrkHeap<16> > {};

template<class SuperHeap>
class SeqFitCommonWrapperHeap :
public CoalesceHeap<ShrinkBrkHeap<AggressiveDiscardHeap<SuperHeap> > > {};


#ifdef FF_FIFO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<FFHeap<FIFO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef FF_LIFO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<FFHeap<LIFO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef FF_AO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<FFHeap<AO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef NF_FIFO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<NFHeap<FIFO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef NF_LIFO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<NFHeap<LIFO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef NF_AO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<NFHeap<AO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef BF_FIFO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<BFHeap<FIFO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef BF_LIFO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<BFHeap<LIFO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef BF_AO
class TheCustomHeapType :
public CommonWrapper<SeqFitCommonWrapperHeap<BFHeap<AO, SeqFitCommonBaseHeap> > > {};
#endif

#ifdef LEAHEAP
class TheCustomHeapType :
public CommonWrapper<LeaHeap<SlopHeap<ChunkHeap<65536, MmapHeap>, 32>, MmapHeap> > {};
#endif

#ifdef DLMALLOC
#include "leamallocheap.h"

class DLMallocHeap :
public LeaMallocHeap {
  public:
	//to avoid ANSIWrapper assertion failure
	inline size_t getSize(void * ptr) {
		size_t size = LeaMallocHeap::getSize(ptr);
		size = size / sizeof(double) * sizeof(double);
		return size;
	}
};

class TheCustomHeapType :
public CommonWrapper<DLMallocHeap> {};
#endif	//DLMALLOC

#ifdef PHKMALLOC
#include "../phkmalloc/phkmallocheap.h"

class TheCustomHeapType :
public CommonWrapper<PhkMallocHeap> {};
#endif	//PHKMALLOC


#ifdef TIMEINTERVAL

#include "leamallocheap.h"
class DLMallocHeap :
public LeaMallocHeap {
  public:
	//to avoid ANSIWrapper assertion failure
	inline size_t getSize(void * ptr) {
		size_t size = LeaMallocHeap::getSize(ptr);
		size = size / sizeof(double) * sizeof(double);
		return size;
	}
};

class TheCustomHeapType :
public ANSIWrapper<TimeIntervalHeap<DLMallocHeap> > {};

#endif	//TIMEINTERVAL

#ifdef VAMalloc

#include "../vam4/vam.h"

class TheCustomHeapType :
public CommonWrapper<SegSizeHeap<128, 496, 32 * 1024, TheOnePageSourceHeap> > {};
#endif	//VAMalloc

/*
#ifdef XALLOC
class TheCustomHeapType :
public ANSIWrapper<XallocHeap<60*1024*1024, MmapHeap> > {};
//public ANSIWrapper<SizeHeap<XallocHeap<60*1024*1024, MmapHeap> > > {};
#endif	//XALLOC
*/

#ifndef VAMalloc

volatile int anyThreadCreated = 0;

#endif	//VAMalloc

#endif


