#ifndef _MEMUSAGE_H_
#define _MEMUSAGE_H_

#include <new>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <map>
#include <signal.h>

#include "heaplayers.h"

#include "pageinfo.h"
#include "util.h"

#define TRACK_FIRST_TOUCH	1
#define TRACK_MADVISE		1
#define TRACK_PAGE_ACCESS	1
#define LOG_PAGE_ACCESS		1
//#define USE_BINARY			1

#define UNPROTECTED_SIZE	512			// window size of unprotected heap pages

#if TRACK_PAGE_ACCESS && !TRACK_FIRST_TOUCH
#error "hey, you must also choose TRACK_FIRST_TOUCH when you choose TRACK_PAGE_ACCESS!"
#endif

#if TRACK_PAGE_ACCESS && !TRACK_MADVISE
#error "hey, you must also choose TRACK_MADVISE when you choose TRACK_PAGE_ACCESS!"
#endif

#if LOG_PAGE_ACCESS && !TRACK_PAGE_ACCESS
#error "hey, you must also choose TRACK_PAGE_ACCESS when you choose LOG_PAGE_ACCESS!"
#endif

#define MEM_PAGE_FIRST_SBRKED	1
#define MEM_PAGE_FIRST_MMAPED	2
#define MEM_PAGE_SBRKED			3
#define MEM_PAGE_MMAPED			4
#define MEM_PAGE_DISCARDED		5


using namespace HL;

struct MyLess {
	bool operator()(const void * a, const void * b) const {
		return (size_t) a < (size_t) b;
	}
};

class MyHeap : public FreelistHeap<StaticHeap<8388608> > {};

template <class TYPE>
class MyAllocator : public STLAllocator<TYPE, MyHeap> {};

typedef map<const void *, size_t, MyLess, MyAllocator<pair<const void *, int> > > mapType;

mapType mappedPages;


#if !defined(RTLD_NEXT)
#define RTLD_NEXT ((void *) -1)
#endif

// Keep track of all heap memory allocated via sbrk and mmap.

extern "C" {
	typedef void * mmapFunctionType (void * addr, size_t len, int prot, int flags, int fildes, off_t off);
	typedef int munmapFunctionType (void * addr, size_t len);
	typedef void exitFunctionType (int);
	typedef void * sbrkFunctionType (intptr_t);
	typedef int brkFunctionType (void * endds);
#if TRACK_MADVISE
	typedef int madviseFunctionType (void * addr, size_t len, int advice);
#endif
#if TRACK_PAGE_ACCESS || TRACK_MADVISE || TRACK_FIRST_TOUCH
	typedef ssize_t readFunctionType (int fd, void * buf, size_t count);
#endif
	static mmapFunctionType * mmapFn = NULL;
	static munmapFunctionType * munmapFn = NULL;
	static sbrkFunctionType * sbrkFn = NULL;
	static brkFunctionType  * brkFn  = NULL;
#if TRACK_MADVISE
	static madviseFunctionType * madviseFn = NULL;
#endif
#if TRACK_PAGE_ACCESS || TRACK_MADVISE || TRACK_FIRST_TOUCH
	static readFunctionType * readFn = NULL;
#endif
	static int devZeroFd = 0;
	static size_t currentAllocatedMmap = 0;
	static size_t currentAllocatedSbrk = 0;
	static size_t currentMmapDiscarded = 0;
	static size_t currentSbrkDiscarded = 0;
#if TRACK_FIRST_TOUCH
	static int currentMmapUntouched = 0;
	static int currentSbrkUntouched = 0;
#endif	// TRACK_FIRST_TOUCH
	static size_t maxAllocated = 0;
	static char * initialBreak = 0;
	static char * currentBreak = 0;
	bool instrument = true;
}

#if TRACK_PAGE_ACCESS

void pageaccess_output(char tag, void * page) {
#if LOG_PAGE_ACCESS
	static int pageaccess_bin = open("pageaccess.bin", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
	assert(pageaccess_bin > 0);

	size_t addr = reinterpret_cast<size_t>(page);
	assert((addr & ~PAGE_MASK) == 0);

	addr += tag;

	write(pageaccess_bin, &addr, sizeof(size_t));
#endif	// LOG_PAGE_ACCESS
}

void pageaccess_printf(const char * fmt, ...) {
#if LOG_PAGE_ACCESS
	static int pageaccess_log = open("pageaccess.log", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
	if (pageaccess_log < 0) {
		abort();
	}

	char buf[100];
	va_list ap;
	va_start(ap, fmt);

	int n = vsnprintf(buf, 100, fmt, ap);
	if (n < 0 || n >= 100) {
		abort();
	}

	write(pageaccess_log, buf, n);
#endif	// LOG_PAGE_ACCESS
}

#include <list>

struct MemRegion {
	size_t start;
	size_t size;
};

class MyListHeap : public FreelistHeap<StaticHeap<1048576> > {};

typedef STLAllocator<MemRegion, MyListHeap> MyListAllocator;

list<MemRegion, MyListAllocator> mappedRegions;

size_t numPages = 0;
size_t numAccessedPages = 0;
size_t numRecentUnmappedPages = 0;
size_t numDiscardedPagesAccessed = 0;

void * unprotectedWindow[UNPROTECTED_SIZE];		// a FIFO window of currently unprotected pages
int windowPointer = 0;
volatile bool newPageAccessed = false;

// reset pages in all memory regions to protected state
void resetAllPages() {
	numPages = 0;
	numAccessedPages = 0;
	numRecentUnmappedPages = 0;
	numDiscardedPagesAccessed = 0;

	list<MemRegion>::iterator r = mappedRegions.begin();
	while (r != mappedRegions.end()) {
		void * addr = reinterpret_cast<void *>(r->start);
		size_t len = r->size;

		assert((reinterpret_cast<size_t>(addr) & ~PAGE_MASK) == 0);
		assert((len & ~PAGE_MASK) == 0);

		numPages += len >> PAGE_SHIFT;

		assert(false);
		int rc = mprotect(addr, len, PROT_NONE);
		assert(rc == 0);

		r++;
	}
}

#endif	// TRACK_PAGE_ACCESS

#if TRACK_MADVISE || TRACK_PAGE_ACCESS
bool pageaccess_initialized = false;

void sigsegvHandler(int sig_num, siginfo_t * info, void * p) {
	int rc;
	void * page = (void *)((size_t)info->si_addr & PAGE_MASK);
//	dbprintf("SIGSEGV on page %p numAccessedPages=%d numPages=%d numRecentUnmappedPages=%d\n", page, numAccessedPages, numPages, numRecentUnmappedPages);

	mapType::iterator i = mappedPages.find(page);

	if (i == mappedPages.end()) {
		dbprintf("SIGSEGV on page %p not found\n", page);
		assert(false);
		abort();
	}

	bool is_discarded = false;

#if TRACK_MADVISE
	if ((*i).second == MEM_PAGE_DISCARDED) {
		int isSbrked = (char *)page >= initialBreak && (char *)page < currentBreak;

		if (isSbrked) {
			(*i).second = MEM_PAGE_SBRKED;
			currentSbrkDiscarded -= PAGE_SIZE;
		}
		else {
			(*i).second = MEM_PAGE_MMAPED;
			currentMmapDiscarded -= PAGE_SIZE;
		}

#if TRACK_PAGE_ACCESS
		numDiscardedPagesAccessed++;
#if USE_BINARY
		pageaccess_output('d', page);
#else
		pageaccess_printf("d %p\n", page);
#endif	// USE_BINARY
#endif	// TRACK_PAGE_ACCESS

		is_discarded = true;
//		dbprintf("SIGSEGV on (%s) discarded page %p\n", isSbrked ? "sbrked" : "mmaped", page);
	}
#endif	// TRACK_MADVISE

	bool is_first_touch_protected = false;

#if TRACK_FIRST_TOUCH
	// if the page is touched for the first time, exclude it from the untouched set
	if (!is_discarded) {
		switch ((*i).second) {
		case MEM_PAGE_FIRST_SBRKED:
#if !TRACK_PAGE_ACCESS		// the page state resetting below will be taken care of if TRACK_PAGE_ACCESS is set
			(*i).second = MEM_PAGE_SBRKED;
#endif	// !TRACK_PAGE_ACCESS
			currentSbrkUntouched -= PAGE_SIZE;
			assert(currentSbrkUntouched >= 0);
			is_first_touch_protected = true;
			break;
		case MEM_PAGE_FIRST_MMAPED:
#if !TRACK_PAGE_ACCESS		// the page state resetting below will be taken care of if TRACK_PAGE_ACCESS is set
			(*i).second = MEM_PAGE_MMAPED;
#endif	// !TRACK_PAGE_ACCESS
			currentMmapUntouched -= PAGE_SIZE;
			assert(currentMmapUntouched >= 0);
			is_first_touch_protected = true;
			break;
		default:
			break;
		}
	}
#endif	// TRACK_FIRST_TOUCH

#if !TRACK_PAGE_ACCESS
	if (!is_discarded && !is_first_touch_protected) {
		dbprintf("SIGSEGV on page %p %d\n", page, (*i).second);

		assert(false);
		abort();
	}
#else	// TRACK_PAGE_ACCESS
	if (!is_discarded) {
		switch ((*i).second) {
		case MEM_PAGE_FIRST_SBRKED:
#if USE_BINARY
			pageaccess_output('S', page);
#else
			pageaccess_printf("S %p\n", page);
#endif
			(*i).second = MEM_PAGE_SBRKED;
			break;
		case MEM_PAGE_SBRKED:
#if USE_BINARY
			pageaccess_output('s', page);
#else
			pageaccess_printf("s %p\n", page);
#endif
			break;
		case MEM_PAGE_FIRST_MMAPED:
#if USE_BINARY
			pageaccess_output('M', page);
#else
			pageaccess_printf("M %p\n", page);
#endif
			(*i).second = MEM_PAGE_MMAPED;
			break;
		case MEM_PAGE_MMAPED:
#if USE_BINARY
			pageaccess_output('m', page);
#else
			pageaccess_printf("m %p\n", page);
#endif
			break;
		default:
			assert(false);
		}
	}

	numAccessedPages++;
//	assert(numAccessedPages - numDiscardedPagesAccessed <= numPages + numRecentUnmappedPages);

	assert(windowPointer >= 0 && windowPointer < UNPROTECTED_SIZE);
	void * old = unprotectedWindow[windowPointer];

	// add this page to unprotected window and remove an old one
	if (old != NULL) {
		i = mappedPages.find(old);
		if (i != mappedPages.end()) {
//			dbprintf("windowPointer=%d old=%p protected again\n", windowPointer, old);
			rc = mprotect(old, PAGE_SIZE, PROT_NONE);
			assert(rc == 0);
		}
		else {
			dbprintf("windowPointer=%d old=%p already removed\n", windowPointer, old);
		}
	}
	unprotectedWindow[windowPointer] = page;
	windowPointer++;
	if (windowPointer == UNPROTECTED_SIZE) {
		windowPointer = 0;
	}
	newPageAccessed = true;

/*
	if (numAccessedPages > 256) {
		resetAllPages();

		rc = mprotect(page, PAGE_SIZE, PROT_READ | PROT_WRITE);
		assert(rc == 0);

		numAccessedPages++;
		assert(numAccessedPages <= numPages);
	}
*/
#endif	// TRACK_PAGE_ACCESS

	rc = mprotect(page, PAGE_SIZE, PROT_READ | PROT_WRITE);
	assert(rc == 0);
}

#endif	// TRACK_MADVISE || TRACK_PAGE_ACCESS

extern "C" {
#if TRACK_MADVISE
	int madvise(void * addr, size_t len, int advice) {
		if (!madviseFn) {
			madviseFn = (madviseFunctionType *) dlsym (RTLD_NEXT, "madvise");
		}

		assert((reinterpret_cast<size_t>(addr) & ~PAGE_MASK) == 0);

		int rc = (*madviseFn)(addr, len, advice);
		assert(rc == 0);

		if (advice == MADV_DONTNEED) {
			if (instrument && rc == 0) {
				if (!pageaccess_initialized) {
					// set SIGSEGV handler
					struct sigaction sa;
					sa.sa_sigaction = sigsegvHandler;
					sa.sa_flags = SA_SIGINFO;

					int rc = sigaction(SIGSEGV, &sa, NULL);
					assert(rc == 0);

#if TRACK_PAGE_ACCESS
					for (int i = 0; i < UNPROTECTED_SIZE; i++) {
						unprotectedWindow[i] = NULL;
					}
#endif	// TRACK_PAGE_ACCESS

					pageaccess_initialized = true;
				}

				dbprintf("original madvise call: %p, %x, %d\n", addr, len, advice);
				size_t advised = 0;
				while (advised < len) {
					void * page = (char *) addr + advised;

					mapType::iterator i = mappedPages.find(page);
					assert(i != mappedPages.end());

//					dbprintf("madvise page %p old status %d\n", page, (*i).second);
#if TRACK_PAGE_ACCESS
#if USE_BINARY
					pageaccess_output('D', page);
#else
					pageaccess_printf("D %p\n", page);
#endif	// USE_BINARY
#endif	// TRACK_PAGE_ACCESS

					switch ((*i).second) {
					case MEM_PAGE_FIRST_SBRKED:
#if TRACK_FIRST_TOUCH
						currentSbrkUntouched -= PAGE_SIZE;
						assert(currentSbrkUntouched >= 0);
#endif	// TRACK_FIRST_TOUCH
					case MEM_PAGE_SBRKED:
						(*i).second = MEM_PAGE_DISCARDED;
						currentSbrkDiscarded += PAGE_SIZE;
						break;
					case MEM_PAGE_FIRST_MMAPED:
#if TRACK_FIRST_TOUCH
						currentMmapUntouched -= PAGE_SIZE;
						assert(currentMmapUntouched >= 0);
#endif	// TRACK_FIRST_TOUCH
					case MEM_PAGE_MMAPED:
						(*i).second = MEM_PAGE_DISCARDED;
						currentMmapDiscarded += PAGE_SIZE;
						break;
					case MEM_PAGE_DISCARDED:
						// page is discarded for multiple times - this is allowed
						break;
					default:
						assert(false);
					}
					assert((*i).second == MEM_PAGE_DISCARDED);

//					dbprintf("madvise page %p new status %d, mprotecting it...\n", page, (*i).second);
					// mprotect the page so we can redo the accouting when future access traps into SIGSEGV handler
					rc = mprotect(page, PAGE_SIZE, PROT_NONE);
					assert(rc == 0);

					advised += PAGE_SIZE;
				}
			}
		}
		return rc;
	}
#endif	// TRACK_MADVISE

#if TRACK_PAGE_ACCESS || TRACK_MADVISE || TRACK_FIRST_TOUCH
	ssize_t read(int fd, void * buf, size_t count) {
		if (!readFn) {
			readFn = (readFunctionType *) dlsym (RTLD_NEXT, "read");
		}

		size_t page;
		size_t start_page = reinterpret_cast<size_t>(buf) & PAGE_MASK;
		size_t end_page = (reinterpret_cast<size_t>(buf) + count + PAGE_SIZE - 1) & PAGE_MASK;
		assert(start_page < end_page);

#if TRACK_PAGE_ACCESS
		// find and remove these pages of the buffer in unprotected window
		for (int i = 0; i < UNPROTECTED_SIZE; i++) {
			page = reinterpret_cast<size_t>(unprotectedWindow[i]);
			assert((page & ~PAGE_MASK) == 0);

			if (page >= start_page && page < end_page) {
				unprotectedWindow[i] = NULL;
			}
		}
#endif	// TRACK_PAGE_ACCESS

		bool buf_in_heap = true;

		page = start_page;
		while (page < end_page) {
			void * page_ptr = reinterpret_cast<void *>(page);
			mapType::iterator i = mappedPages.find(page_ptr);
			if (i != mappedPages.end()) {	// is a heap page
				switch ((*i).second) {
				case MEM_PAGE_FIRST_SBRKED:
#if TRACK_PAGE_ACCESS
#if USE_BINARY
					pageaccess_output('S', page_ptr);
#else
					pageaccess_printf("S %p\n", page_ptr);
#endif
#endif	// TRACK_PAGE_ACCESS
#if TRACK_FIRST_TOUCH
					currentSbrkUntouched -= PAGE_SIZE;
					assert(currentSbrkUntouched >= 0);
#endif	// TRACK_FIRST_TOUCH
					(*i).second = MEM_PAGE_SBRKED;
					break;
				case MEM_PAGE_SBRKED:
#if TRACK_PAGE_ACCESS
#if USE_BINARY
					pageaccess_output('s', page_ptr);
#else
					pageaccess_printf("s %p\n", page_ptr);
#endif
#endif	// TRACK_PAGE_ACCESS
					break;
				case MEM_PAGE_FIRST_MMAPED:
#if TRACK_PAGE_ACCESS
#if USE_BINARY
					pageaccess_output('M', page_ptr);
#else
					pageaccess_printf("M %p\n", page_ptr);
#endif
#endif	// TRACK_PAGE_ACCESS
#if TRACK_FIRST_TOUCH
					currentMmapUntouched -= PAGE_SIZE;
					assert(currentMmapUntouched >= 0);
#endif	// TRACK_FIRST_TOUCH
					(*i).second = MEM_PAGE_MMAPED;
					break;
				case MEM_PAGE_MMAPED:
#if TRACK_PAGE_ACCESS
#if USE_BINARY
					pageaccess_output('m', page_ptr);
#else
					pageaccess_printf("m %p\n", page_ptr);
#endif
#endif	// TRACK_PAGE_ACCESS
					break;
				default:
					assert(false);
				}

				int rc = mprotect(page_ptr, PAGE_SIZE, PROT_READ | PROT_WRITE);
				assert(rc == 0);
			}
			else {
				assert(page == start_page);
				buf_in_heap = false;
				break;
			}

			page += PAGE_SIZE;
		}

		ssize_t num_bytes = (*readFn)(fd, buf, count);
		dbprintf("read(%d, %p, %d) returns %d\n", fd, buf, count, num_bytes);

#if TRACK_PAGE_ACCESS
		if (buf_in_heap) {
			// protect these pages again
			page = start_page;
			while (page < end_page) {
				int rc = mprotect(reinterpret_cast<void *>(page), PAGE_SIZE, PROT_NONE);
				assert(rc == 0);

				page += PAGE_SIZE;
			}
		}
#endif	// TRACK_PAGE_ACCESS

		return num_bytes;
	}
#endif	// TRACK_PAGE_ACCESS || TRACK_MADVISE || TRACK_FIRST_TOUCH

	void * mmap(void * addr, size_t len, int prot, int flags, int fildes, off_t off) {
		if (!mmapFn) {
			devZeroFd = open ("/dev/zero", O_RDWR);
			mmapFn = (mmapFunctionType *) dlsym (RTLD_NEXT, "mmap");
		}
		// Check to see if these are heap pages (anonymous or /dev/zero).
		if (((flags & MAP_ANON) && (fildes == -1)) || (fildes == devZeroFd)) {
			if (instrument) {
				void * ptr = (*mmapFn)(addr, len, prot, flags, fildes, off);
				dbprintf("original mmap call: %x, %x\n", ptr, len);
				assert((reinterpret_cast<size_t>(ptr) & ~PAGE_MASK) == 0);
				size_t mapped = 0;
				while (mapped < len) {
					mappedPages[(char *) ptr + mapped] = MEM_PAGE_FIRST_MMAPED;
					dbprintf("mmap page %p\n", (char *) ptr + mapped);
					mapped += PAGE_SIZE;
					currentAllocatedMmap += PAGE_SIZE;
				}
				if (currentAllocatedMmap + currentAllocatedSbrk > maxAllocated) {
					maxAllocated = currentAllocatedMmap + currentAllocatedSbrk;
				}
#if TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH
				if (!pageaccess_initialized) {
					// set SIGSEGV handler
					struct sigaction sa;
					sa.sa_sigaction = sigsegvHandler;
					sa.sa_flags = SA_SIGINFO;

					int rc = sigaction(SIGSEGV, &sa, NULL);
					assert(rc == 0);

#if TRACK_PAGE_ACCESS
					for (int i = 0; i < UNPROTECTED_SIZE; i++) {
						unprotectedWindow[i] = NULL;
					}
#endif	// TRACK_PAGE_ACCESS

					pageaccess_initialized = true;
				}
#endif	// TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH

#if TRACK_PAGE_ACCESS
				MemRegion new_region;
				new_region.start = reinterpret_cast<size_t>(ptr);
				new_region.size = (len + PAGE_SIZE - 1) & PAGE_MASK;
				mappedRegions.push_front(new_region);

				numPages += (len + PAGE_SIZE - 1) >> PAGE_SHIFT;
#endif	// TRACK_PAGE_ACCESS

#if TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH
				currentMmapUntouched += (len + PAGE_SIZE - 1) & PAGE_MASK;
				assert(currentMmapUntouched <= currentAllocatedMmap && (currentMmapUntouched & ~PAGE_MASK) == 0);
				dbprintf("mmap mprotecting ptr=%p len=%x\n", ptr, len);

				int rc = mprotect(ptr, len, PROT_NONE);
				assert(rc == 0);
#endif	// TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH

				return (void *) ptr;
			}
		}
		return (*mmapFn)(addr, len, prot, flags, fildes, off);
	}

	int munmap(void * addr, size_t len) {
		if (!munmapFn) {
			munmapFn = (munmapFunctionType *) dlsym (RTLD_NEXT, "munmap");
		}
		if (instrument) {
			dbprintf("original munmap %p, %x\n", addr, len);
			assert((reinterpret_cast<size_t>(addr) & ~PAGE_MASK) == 0);

			size_t unmapped = 0;
			while (unmapped < len) {
				void * page = (char *)addr + unmapped;

				mapType::iterator i = mappedPages.find (page);
				assert(i != mappedPages.end());

				dbprintf("munmap page %p status %d\n", page, (*i).second);
				if ((*i).second == MEM_PAGE_DISCARDED) {
					currentMmapDiscarded -= PAGE_SIZE;
				}
				else if ((*i).second == MEM_PAGE_FIRST_MMAPED) {
#if TRACK_FIRST_TOUCH
					currentMmapUntouched -= PAGE_SIZE;
					assert(currentMmapUntouched >= 0);
#endif	// TRACK_FIRST_TOUCH
				}
				else {
					assert((*i).second == MEM_PAGE_MMAPED);
				}

				mappedPages.erase (i);
				unmapped += PAGE_SIZE;
				currentAllocatedMmap -= PAGE_SIZE;
#if TRACK_PAGE_ACCESS
#if USE_BINARY
				pageaccess_output('U', page);
#else
				pageaccess_printf("U %p\n", page);
#endif

				// remove the page from mappedRegions
				list<MemRegion>::iterator r = mappedRegions.begin();
				while (r != mappedRegions.end()) {
					if (reinterpret_cast<size_t>(page) >= r->start && reinterpret_cast<size_t>(page) < r->start + r->size) {
						if (reinterpret_cast<size_t>(page) == r->start) {	// first page of memory region
							r->start += PAGE_SIZE;
							r->size -= PAGE_SIZE;
							if (r->size == 0) {
								mappedRegions.erase(r);
							}
						}
						else if (reinterpret_cast<size_t>(page) == r->start + r->size - PAGE_SIZE) {	// last page of memory region
							r->size -= PAGE_SIZE;
							assert(r->size != 0);
						}
						else {		// in the middle, split the region
							MemRegion new_region;
							new_region.start = reinterpret_cast<size_t>(page) + PAGE_SIZE;
							new_region.size = r->start + r->size - new_region.start;

							r->size = reinterpret_cast<size_t>(page) - r->start;

							mappedRegions.push_front(new_region);
						}
						break;
					}
					r++;
				}
				assert(r != mappedRegions.end());

				numPages--;
				numRecentUnmappedPages++;
#endif
			}
		}
		return (*munmapFn) (addr, len);
	}

	void addPagesSbrked(size_t start, size_t len) {
		assert((start & ~PAGE_MASK) == 0);
		assert((len & ~PAGE_MASK) == 0);
		size_t offset = 0;
		while (offset < len) {
			void * page = reinterpret_cast<void *>(start + offset);
			mapType::iterator i = mappedPages.find(page);
			assert(i == mappedPages.end());

			mappedPages[page] = MEM_PAGE_FIRST_SBRKED;
//			dbprintf("add page sbrked %p currentAllocatedSbrk=%x\n", page, currentAllocatedSbrk);

			offset += PAGE_SIZE;
		}
#if TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH
		if (!pageaccess_initialized) {
			// set SIGSEGV handler
			struct sigaction sa;
			sa.sa_sigaction = sigsegvHandler;
			sa.sa_flags = SA_SIGINFO;

			int rc = sigaction(SIGSEGV, &sa, NULL);
			assert(rc == 0);

#if TRACK_PAGE_ACCESS
			for (int i = 0; i < UNPROTECTED_SIZE; i++) {
				unprotectedWindow[i] = NULL;
			}
#endif	// TRACK_PAGE_ACCESS

			pageaccess_initialized = true;
		}
#endif	// TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH

#if TRACK_PAGE_ACCESS
		static bool first_one = true;
		if (first_one) {
			MemRegion new_region;
			new_region.start = start;
			new_region.size = len;

			mappedRegions.push_front(new_region);

			first_one = false;
		}
		else {
			list<MemRegion>::iterator r = mappedRegions.begin();
			while (r != mappedRegions.end()) {		// search the sbrk region
				if (start == r->start + r->size) {	// new region must be attached to the end of the existing sbrk region
					r->size += len;
					break;
				}
				r++;
			}
			assert(r != mappedRegions.end());
			if (r == mappedRegions.end()) {
				abort();
			}
		}
		numPages += len >> PAGE_SHIFT;
#endif	// TRACK_PAGE_ACCESS

#if TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH
		currentSbrkUntouched += len;
		dbprintf("currentAllocatedSbrk=%x currentSbrkUntouched=%x initialBreak=%p currentBreak=%p\n", currentAllocatedSbrk, currentSbrkUntouched, initialBreak, currentBreak);
		assert(currentSbrkUntouched <= currentAllocatedSbrk + PAGE_SIZE && (currentSbrkUntouched & ~PAGE_MASK) == 0);

		dbprintf("addPagesSbrked mprotecting start=%p len=%x\n", start, len);
		int rc = mprotect(reinterpret_cast<void *>(start), len, PROT_NONE);
		assert(rc == 0);
#endif	// TRACK_PAGE_ACCESS || TRACK_FIRST_TOUCH

	}

	void removePagesSbrked(size_t start, size_t len) {
		assert((start & ~PAGE_MASK) == 0);
		assert((len & ~PAGE_MASK) == 0);
		size_t offset = 0;
		while (offset < len) {
			void * page = reinterpret_cast<void *>(start + offset);
			mapType::iterator i = mappedPages.find(page);
			assert(i != mappedPages.end());

			dbprintf("remove page sbrked %p status %d\n", page, (*i).second);
			if ((*i).second == MEM_PAGE_DISCARDED) {
				currentSbrkDiscarded -= PAGE_SIZE;
			}
			else if ((*i).second == MEM_PAGE_FIRST_SBRKED) {
#if TRACK_FIRST_TOUCH
				currentSbrkUntouched -= PAGE_SIZE;
				assert(currentSbrkUntouched >= 0);
#endif	// TRACK_FIRST_TOUCH
			}
			else {
				assert((*i).second == MEM_PAGE_SBRKED);
			}

			mappedPages.erase(i);
			offset += PAGE_SIZE;
#if TRACK_PAGE_ACCESS
#if USE_BINARY
			pageaccess_output('U', page);
#else
			pageaccess_printf("U %p\n", page);
#endif
#endif	// TRACK_PAGE_ACCESS
		}
#if TRACK_PAGE_ACCESS
		list<MemRegion>::iterator r = mappedRegions.begin();
		while (r != mappedRegions.end()) {		// search the sbrk region
			if (start + len == r->start + r->size) {	// region must be detached from the end of existing sbrk region
				assert(len <= r->size);
				r->size -= len;

				numPages -= len >> PAGE_SHIFT;
				numRecentUnmappedPages += len >> PAGE_SHIFT;

				if (r->size == 0) {
					mappedRegions.erase(r);
					assert(false);
					abort();
				}
				break;
			}
			r++;
		}
		assert(r != mappedRegions.end());
		if (r == mappedRegions.end()) {
			abort();
		}
#endif
	}

	void * sbrk (intptr_t incr) {
		if (!sbrkFn) {
			brkFn = (brkFunctionType *) dlsym (RTLD_NEXT, "brk");
			sbrkFn = (sbrkFunctionType *) dlsym (RTLD_NEXT, "sbrk");
			initialBreak = (char *) (*sbrkFn)(0);
			currentBreak = initialBreak;
		}

		void * p = (*sbrkFn)(incr);
		assert(p == currentBreak);
		if (instrument && p == currentBreak) {
			dbprintf("original sbrk(%d) returns %p\n", incr, p);
			assert(currentBreak + incr >= initialBreak);

			size_t old_page_end = ((size_t) currentBreak + PAGE_SIZE - 1) & PAGE_MASK;

			currentBreak += incr;
			assert(currentBreak == (*sbrkFn)(0));

			size_t new_page_end = ((size_t) currentBreak + PAGE_SIZE - 1) & PAGE_MASK;

			currentAllocatedSbrk = currentBreak - initialBreak;
			if (currentAllocatedMmap + currentAllocatedSbrk > maxAllocated) {
				maxAllocated = currentAllocatedMmap + currentAllocatedSbrk;
			}

			dbprintf("old_page_end=%p new_page_end=%p\n", old_page_end, new_page_end);
			if (new_page_end > old_page_end) {
				assert(incr > 0);
				addPagesSbrked(old_page_end, new_page_end - old_page_end);
			}
			else if (new_page_end < old_page_end) {
				assert(incr < 0);
				removePagesSbrked(new_page_end, old_page_end - new_page_end);
			}
		}
		return p;
	}

	int brk(void * endds) {
		if (!brkFn) {
			brkFn = (brkFunctionType *) dlsym (RTLD_NEXT, "brk");
			sbrkFn = (sbrkFunctionType *) dlsym (RTLD_NEXT, "sbrk");
			initialBreak = (char *) (*sbrkFn)(0);
			currentBreak = initialBreak;
		}

		int rc = (*brkFn)(endds);
		assert(rc == 0);

		if (instrument && rc == 0) {
			dbprintf("original brk = %p\n", endds);
			assert((char *) endds >= initialBreak);

			size_t old_page_end = ((size_t) currentBreak + PAGE_SIZE - 1) & PAGE_MASK;

			currentBreak = (char *)endds;
			assert(currentBreak == (*sbrkFn)(0));

			size_t new_page_end = ((size_t) currentBreak + PAGE_SIZE - 1) & PAGE_MASK;

			currentAllocatedSbrk = currentBreak - initialBreak;
			if (currentAllocatedMmap + currentAllocatedSbrk > maxAllocated) {
				maxAllocated = currentAllocatedMmap + currentAllocatedSbrk;
			}

			dbprintf("old_page_end=%p new_page_end=%p\n", old_page_end, new_page_end);
			if (new_page_end > old_page_end) {
				addPagesSbrked(old_page_end, new_page_end - old_page_end);
			}
			else if (new_page_end < old_page_end) {
				removePagesSbrked(new_page_end, old_page_end - new_page_end);
			}
		}
		return rc;
	}

	void exit(int status) {
		instrument = false;
		exitFunctionType * fn = (exitFunctionType *) dlsym (RTLD_NEXT, "exit");
		dbprintf("Current allocated with mmap = %d\n", currentAllocatedMmap);
		dbprintf("Current allocated with sbrk = %d\n", currentAllocatedSbrk);
		dbprintf("Max allocated overall = %d\n", maxAllocated);
		if (fn) {
			(*fn)(status);
		}
	}
}

#endif
