#ifndef _TIMEINTERVALHEAP_H
#define _TIMEINTERVALHEAP_H_

#include <stdio.h>
#include <stdarg.h>

#include "timer.h"

#define INTERVAL		0.1
#define CHECK_FREQUENCY	100


namespace HL {

template<class SuperHeap>
class TimeIntervalHeap : public SuperHeap {
  public:
	TimeIntervalHeap() : num_malloc_free(0), num_log(0) {
		getTime(start_low, start_high);
	}

	~TimeIntervalHeap() {
		ti_printf("time elapsed: %f seconds\n", getTimeElapsed());
	}


	inline void * malloc(size_t size) {
		if (num_malloc_free++ % CHECK_FREQUENCY == 0) {
			double elapsed = getTimeElapsed();
			if (elapsed >= INTERVAL * num_log) {
				ti_printf("T %d %f %f\n", num_malloc_free, INTERVAL * num_log, elapsed);
				num_log++;

				if (elapsed >= INTERVAL * num_log) {
					//ti_printf("oops!\n");
				}
			}
		}

		return SuperHeap::malloc(size);
	}

	inline void free(void * ptr) {
		if (num_malloc_free++ % CHECK_FREQUENCY == 0) {
			double elapsed = getTimeElapsed();
			if (elapsed >= INTERVAL * num_log) {
				ti_printf("T %d %f %f\n", num_malloc_free, INTERVAL * num_log, elapsed);
				num_log++;

				if (elapsed >= INTERVAL * num_log) {
					//ti_printf("oops!\n");
				}
			}
		}

		SuperHeap::free(ptr);
	}

  private:
	unsigned long start_low, start_high;
	unsigned long current_low, current_high;

	size_t num_malloc_free;
	size_t num_log;

	double getTimeElapsed() {
		getTime(current_low, current_high);
		double d = (double) (current_low - start_low) + (double) (UINT_MAX)*(double)(current_high - start_high);

		static double freq = getFrequency();

		return d / freq;
	}

	void ti_printf(const char * fmt, ...) {
		static int ti_log = open("ti.log", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
		if (ti_log < 0) {
			abort();
		}

		char buf[200];
		va_list ap;
		va_start(ap, fmt);

		int n = vsnprintf(buf, 200, fmt, ap);
		if (n < 0 || n >= 200) {
			abort();
		}

		write(ti_log, buf, n);
	}


};	//end of class TimeIntervalHeap
};	//end of namespace HL


#endif
