#ifndef _UTIL_H_
#define _UTIL_H_

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#define BUF_SZ 1000

//output debug messages without invoking malloc()
void dbprintf(const char *fmt, ...) {
#ifdef DEBUG

#if DB_PRINT_TO_FILE
	static int debug_log = open("debug.log", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
	if (debug_log < 0) {
		abort();
	}
#endif

	char buf[BUF_SZ];
	va_list ap;
	va_start(ap, fmt);

	int n = vsnprintf(buf, BUF_SZ, fmt, ap);
	if (n < 0 || n >= BUF_SZ) {
		abort();
	}

#if DB_PRINT_TO_FILE
	write(debug_log, buf, n);
#else
	printf("<dbprintf>");
	printf(buf);
#endif

#endif
}

#ifdef MYASSERT
#undef assert

void __assert_fail (__const char *__assertion, __const char *__file,
					unsigned int __line, __const char *__function) {
	dbprintf("ASSERTION FAILED!\n[%s:%d] in function %s:\n(%s) was false!\n", __file, __line, __function, __assertion);
	abort();
}

#define __ASSERT_FUNCTION __PRETTY_FUNCTION__

#define assert(expr) \
  (__ASSERT_VOID_CAST ((expr) ? 0 :					      \
		       (__assert_fail (__STRING(expr), __FILE__, __LINE__,    \
				       __ASSERT_FUNCTION), 0)))

#else

#include <assert.h>

#endif


inline void memClean(void * start, size_t bytes) {
	assert(reinterpret_cast<size_t>(start) % sizeof(size_t) == 0);
	assert(bytes % sizeof(size_t) == 0);

	for (int i = 0; i < bytes / sizeof(size_t); i++) {
		reinterpret_cast<size_t *>(start)[i] = 0;
	}
}

#endif
