#include <stdlib.h>

namespace DL  {

  // 32-bit only for now...
  class sizeBlob {
  public:
    size_t size:30;
    size_t prev_inuse:1;
    size_t is_mmapped:1;
  };

  typedef size_t INTERNAL_SIZE_T;
  typedef unsigned long PTR_UINT;
  typedef unsigned long CHUNK_SIZE_T;

  //  enum { NON_MAIN_ARENA = 0x4 };
  //  enum { SIZE_BITS = (PREV_INUSE | IS_MMAPPED | NON_MAIN_ARENA) };
  enum { MALLOC_ALIGNMENT = 2 * sizeof(size_t) };
  enum { MALLOC_ALIGN_MASK = MALLOC_ALIGNMENT - 1 };

  // EDB FIX ME: This should change to a template expression.
  enum { PAGE_SIZE = 4096 };

  enum { M_MXFAST = 64 };

  inline bool aligned_OK (void * m) {
    return (((PTR_UINT)(m) & MALLOC_ALIGN_MASK) == 0);
  }

  class malloc_chunk {
  private:

    size_t prev_size;
    sizeBlob size_field;
    malloc_chunk * fd;
    malloc_chunk * bk;

  public:

    /// Returns a coalesced malloc_chunk.
    /// @note After executing, 'this' may be invalid! Use the return value instead.
    malloc_chunk * combineAdjacentFree (void) {
      malloc_chunk * new_object = this;
      if (!nextChunk()->is_inuse()) {
	size_field.size += 2 * sizeof(size_t) + nextChunk()->chunksize();
      }
      if (!previousChunk()->is_inuse()) {
	new_object = previousChunk();
	new_object->size_field.size += 2 * sizeof(size_t) + chunksize();
      }
      return new_object;
    }

    malloc_chunk * previousChunk (void) const {
      return mem2chunk ((char *) this - prev_size);
    }

    malloc_chunk * nextChunk (void) const {
      return (malloc_chunk *) ((char *) this + this->size_field.size);
    }

    /// Locate the malloc_chunk for a malloc'ed pointer.
    static malloc_chunk * mem2chunk (void * mem) {
      return ((malloc_chunk *) ((char *) mem - 2 * sizeof(size_t)));
    }

    /// Find the usable memory area given a malloc_chunk.
    inline void * chunk2mem (void) const {
      return ((void *)((char*) this + 2*sizeof(size_t)));
    }

    /// Obtain the usable size of a chunk.
    inline size_t usable_size (void) const {
      if (is_mmapped()) 
	return chunksize() - 2 * sizeof(size_t);
      else if (is_inuse())
	return chunksize() - sizeof(size_t);
      else return 0;
    }

    inline size_t chunksize (void) const {
      return size_field.size;
    }

    inline bool is_mmapped (void) const {
      return size_field.is_mmapped;
    }

    inline bool is_inuse (void) const {
      // Check the next object's header to see if this object is in use.
      return nextChunk()->size_field.prev_inuse;
    }
  };


  enum { MIN_CHUNK_SIZE = sizeof(malloc_chunk) };
  enum { MINSIZE = (unsigned long)(((MIN_CHUNK_SIZE+MALLOC_ALIGN_MASK) & ~MALLOC_ALIGN_MASK))};

  bool REQUEST_OUT_OF_RANGE (size_t req)
  {
    return ((CHUNK_SIZE_T)(req) >=
	    (CHUNK_SIZE_T)(INTERNAL_SIZE_T)(-2 * MINSIZE));
  }

}
